<?php

namespace Drupal\field_history\Controller;

use Drupal\Component\Utility\Html;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\node\Entity\Node;
use Drupal\node\NodeInterface;

class FieldHistoryTabController extends ControllerBase {

  // Given a node ID, returns an array of changes to that field keyed by revision ID:
  // [
  //   1 => [['value' => 'foo']],
  //   2 => [['value' => 'bar'], ['value' => 'foo']],
  //   'current' => [['value' => 'baz'], ['value' => 'bar'], ['value' => 'baz']]
  // ]
  public static function getFieldHistory(NodeInterface $node, string $field) {
    $current_values = $node->get($field)->getValue();
    $return = [];
    // First just get all the revisions, ordered by change date:
    if ($node->get($field)->getFieldDefinition()->getFieldStorageDefinition()->isBaseField()) {
      $query = \Drupal::database()
        ->select("node_field_revision", 'field')
        ->orderBy('field.changed')
        ->condition('field.nid', $node->id());
      $query->addField('field', 'vid', 'revision_id');
    } else {
      $query = \Drupal::database()
        ->select("node_revision__{$field}", 'field')
        ->fields('field', ['revision_id'])
        ->condition('entity_id', $node->id());
      $query->join('node_field_revision', 'revision', 'revision.vid = field.revision_id');
      $query->orderBy('revision.changed', 'ASC');
    }
    $results = $query->execute()->fetchAll(\PDO::FETCH_ASSOC);
    $previous_value = [];
    foreach ($results as $result) {
      $revision = \Drupal::entityTypeManager()->getStorage('node')->loadRevision($result['revision_id']);
      $revision_value = $revision->get($field)->getValue();
      if (serialize($revision_value) == serialize($previous_value)) continue;
      $return[$result['revision_id']] = $revision_value;
      $previous_value = $revision_value;
    }
    ksort($return);
    $return['current'] = $current_values;
    return $return;
  }

  public function fieldHistory(NodeInterface $node, string $field, $output = '') {
    $history = static::getFieldHistory($node, $field);
    $field_label = $node->get($field)->getFieldDefinition()->getLabel();
    $return = [
      'breadcrumb' => [
        '#theme' => 'breadcrumb',
        '#links' => [
          Link::createFromRoute($node->getTitle(), 'entity.node.canonical', ['node' => $node->id()]),
          Link::createFromRoute('Field History', 'entity.node.field_history', ['node' => $node->id()]),
          Link::createFromRoute($field_label, 'entity.node.field_history', ['node' => $node->id(), 'field' => $field]),
        ]
      ],
      'title' => [
        '#theme' => 'page_title',
        '#title' => '"' . $field_label . '" Field History for ' . $node->getTitle()
      ],
      'output' => [
        '#type' => 'container',
        'link' => [
          '#type' => 'link'
        ]
      ]
    ];
    if ('raw' === $output) {
      $return['output']['link']['#title'] = 'View Rendered Items';
      $return['output']['link']['#url'] = Url::fromRoute('entity.node.field_history', ['node' => $node->id(), 'field' => $field, 'output' => '']);
    } else {
      $return['output']['link']['#title'] = 'View Raw Values';
      $return['output']['link']['#url'] = Url::fromRoute('entity.node.field_history', ['node' => $node->id(), 'field' => $field, 'output' => 'raw']);
    }
    $i = 0;
    foreach ($history as $revision_id => $values) {
      if ('current' === $revision_id) {
        $revision = $node;
        $header = "Current revision ({$node->getRevisionId()})";
      } else {
        $revision = \Drupal::entityTypeManager()->getStorage('node')->loadRevision($revision_id);
        $header = "Revision {$revision_id}";
      }
      $return[$revision_id] = [
        '#type' => 'container',
        'header' => ['#markup' => "<h2>$header</h2>"],
        'time' => [
          '#type' => 'html_tag',
          '#tag' => 'time',
          '#value' => date('F j, Y H:i:s', $revision->getRevisionCreationTime())
        ],
        'value' => $revision->get($field)->view(['label' => 'hidden'])
      ];
      if ($output === 'raw') {
        $items = [];
        foreach ($revision->get($field)->getValue() as $delta => $value) {
          $items[$delta] = [
            '#type' => 'html_tag',
            '#tag' => 'pre',
            '#value' => Html::escape(print_r($value, true)),
          ];
        }
        $return[$revision_id]['value'] = [
          '#theme' => 'item_list',
          '#items' => $items,
          '#type' => 'ol'
        ];
      }
      $i++;
    }
    return $return;
  }

  public function overview(NodeInterface $node) {
    $fields = $node->getFields(FALSE);
    $header = [
      'field_name' => $this->t('Field Name'),
      'current_value' => $this->t('Current Value'),
      'previous_value' => $this->t('Previous Value'),
      'date_changed' => $this->t('Date changed'),
    ];
    $rows = [];
    foreach ($fields as $field) {
      $field_name = $field->getName();
      $storage = $field->getFieldDefinition()->getFieldStorageDefinition();
      if (!$storage->isRevisionable()) continue;
      $row = [
        'field_name' => Link::createFromRoute($field->getFieldDefinition()->getLabel() . " ($field_name)", 'entity.node.field_history', ['node' => $node->id(), 'field' => $field_name]),
        'current_value' => ['data' => $node->get($field_name)->view(['label' => 'hidden'])],
        'previous_value' => [],
        'date_changed' => ''
      ];
      $history = static::getFieldHistory($node, $field_name);
      if (count($history) > 2) { // 2 = initial and current (always present)
        // Ignore the current revision. It may be a duplicate of another revision listed
        $history_filtered = array_filter($history, function ($key) {
          return $key !== 'current';
        }, ARRAY_FILTER_USE_KEY);
        $keys = array_keys($history_filtered);
        $revision_id = $keys[count($history_filtered) - 2];
        $revision_when_changed_id = $keys[count($history_filtered) - 1];
        $revision = \Drupal::entityTypeManager()->getStorage('node')->loadRevision($revision_id);
        $revision_when_changed = \Drupal::entityTypeManager()->getStorage('node')->loadRevision($revision_when_changed_id);
        if ($revision) {
          $row['previous_value']['data'] = $revision->get($field_name)->view(['label' => 'hidden']);
          $row['date_changed'] = date('F j, Y H:i:s', $revision_when_changed->getRevisionCreationTime());
        }

      }


      $rows[] = $row;
    }
    return [
      'table' => [
        '#type' => 'table',
        '#header' => $header,
        '#rows' => $rows
      ]
    ];
  }

  public function content(NodeInterface $node, $field = '', $output = '') {
    if ($field) {
      return $this->fieldHistory($node, $field, $output);
    } else {
      return $this->overview($node);
    }
  }

  public function title(NodeInterface $node, $field = '', $output = '') {
    if ($field) {
      return $node->get($field)->getFieldDefinition()->getLabel() . ' Field History for ' . $node->getTitle();
    } else {
      return $node->getTitle() . ' Field History';
    }
  }
}